;****************************************************************************
; DSKWATCH v3.04 - (c) 1987-1993 Vortex Software
;
; Author : Michael J Phillips
; Date   : 27 Oct 91
; Address: 5/5 Waddell Place
;          Curtin ACT 2605
;          Australia
; Tel    : (06) 2811980
;
;   This program monitors the disk interrupt (13h) for any errors.
; When any error occurs, a message is displayed in the top-right
; corner of the screen.
;
;   The idea is based on an article/program written by Steve Holzner,
; and published in the June 11, 1985 PC Magazine.
;
;   I modified the original program several times, to tailor it
; to my needs.  This version (total rewrite) uses several of the
; TSR/multiplex routines that were published in a book called
; "DOS5 Techniques and Utilities".
;****************************************************************************
; 20 Oct 91 v3.00 - Major rewrite.
; 27 Oct 91 v3.01 - Fixed hex conversion bug in HEX2
; 04 Feb 93 v3.02 - Fixed bug in display of "Unknown error" message
; 05 Feb 93 v3.03 - Changed the length equates to use "$-"
; 09 Feb 93 v3.04 - Simplified the logic for MPLEX_INT
;****************************************************************************

;****************************************************************************
; MASK_ALHI (MACRO)
;   Return the high nibble of AL, in the lo nibble of AL.
;****************************************************************************
mask_alhi       macro
                and     al, 0F0h
                mov     cl, 4
                shr     al, cl
                endm

;****************************************************************************
; MASK_ALLO (MACRO)
;   Return the lo nibble of AL.
;****************************************************************************
mask_allo       macro
                and     al, 0Fh
                endm

;****************************************************************************
; HEX2ASC (MACRO)
;   Convert the lo nibble of AL to an ASCII character equivalent.
;****************************************************************************
hex2asc         macro
                local   less10
                cmp     al, 10
                jb      less10
                add     al, 'A'-'0'-10
less10:
                add     al, '0'
                endm

LF              equ     10
CR              equ     13

INT10_GET_MODE  equ     0Fh

floppy_drives   equ     01h
hard_drives     equ     02h
both_drives     equ     floppy_drives + hard_drives

code            segment
                assume  cs:code
                org     100h
begin:          jmp     init

Signature       db      "DWv3.04"               ;Program signature/version
SignatureLen    equ     $-Signature
prog_id         db      ?                       ;Multiplex ID number

int13h          dd      ?                       ;interrupt 13H vector
int2Fh          dd      ?                       ;Interrupt 2FH vector

Drive           db      ' xxh: '
DriveLen        equ     $-Drive
;                       Error    Error                      Floppy      Fixed
;                       Code     Description                Disk        Disk
;                       ----  -------------------           ------      ------
DskErr          db      0FFh,'Sense Op. Failed   ';          No          Yes
                db      0E0h,'Status Error       ';          No          Yes
                db      0CCh,'Write Fault        ';          No          Yes
                db      0BBh,'Undefined Error    ';          No          Yes
                db      0AAh,'Drive Not Ready    ';          No          Yes
                db      80h, 'Drive Timed Out    ';          Yes         Yes
                db      40h, 'Seek Op. Failed    ';          Yes         Yes
                db      20h, 'Controller Failed  ';          Yes         Yes
                db      11h, 'ECC Corrected Error';          No          Yes
                db      10h, 'Bad CRC or ECC     ';          Yes         Yes
                db      0Fh, 'DMA Arbitration    ';          No          Yes
                db      0Eh, 'Address Mark Found ';          No          Yes
                db      0Dh, 'Invalid Sector No. ';          No          Yes
                db      0Ch, 'Unknown Media Type ';          Yes         No
                db      0Bh, 'Bad Track Detected ';          No          Yes
                db      0Ah, 'Bad Sector Flag    ';          No          Yes
                db      09h, 'DMA Across Boundary';          Yes         Yes
                db      08h, 'DMA Overrun        ';          Yes         Yes
                db      07h, 'Bad Parameter Table';          No          Yes
                db      06h, 'No Disk in Drive   ';          Yes         No
                db      05h, 'Reset Failed       ';          No          Yes
                db      04h, 'Sector Not Found   ';          Yes         Yes
                db      03h, 'Write Protected    ';          Yes         No
                db      02h, 'No Address Mark    ';          Yes         Yes
                db      01h, 'Bad Command        ';          Yes         Yes
NumMsgs         equ     25
UnkErc          db      00h, 'Unknown Error xxh  '
NextMsg         equ     $-UnkErc
MsgLen          equ     $-UnkErc-1
dw_drives       db      ?                       ;dskwatch drive
                                                ;01 = floppy_drives
                                                ;02 = hard_drives
                                                ;03 = both_drives

RetFlg          dw      ?
RetAddr         dd      ?
VidSeg          equ     0B000h
VidOfs          dw      00000h                  ;set to 8000h for colour


;****************************************************************************
; DISK_INT handles interrupt 13H.
;****************************************************************************

disk_int        proc    far
                pushf                           ;Push FLAGS register
                call    cs:[int13h]             ;Call previous handler
                pushf
                pop     cs:RetFlg
                jnc     disk_exit               ;no error

                cmp     dw_drives, both_drives
                je      disp_err

                cmp     dw_drives, hard_drives
                jne     chk
                cmp     dl, 80h
                jae     disp_err
                jmp     disk_exit

chk:
                cmp     dw_drives, floppy_drives
                jne     disk_exit
                cmp     dl, 80h
                jae     disk_exit
                ;jmp     disp_err
disp_err:
                ;save registers
                push    ax
                push    dx
                push    cx
                push    di
                push    si
                push    es

                call    DspErr

                ;restore registers
                pop     es
                pop     si
                pop     di
                pop     cx
                pop     dx
                pop     ax
disk_exit:
                ;get the return address from the stack
                pop     word ptr cs:RetAddr
                pop     word ptr cs:RetAddr+2
                ;fix up the stack
                add     sp, 2
                ;restore the flags to their original value
                push    cs:RetFlg
                popf
                ;simulate a return from interrupt
                jmp     dword ptr cs:RetAddr
disk_int        endp

DspErr          proc    near
                ;set the drive number for the error message
                lea     si, cs:Drive+1
                mov     al, dl
                call    Hex2

                ;make ES:DI point to screen address
                mov     dx, VidSeg
                mov     es, dx
                mov     di, cs:VidOfs

                ;make CS:SI point to error message
                lea     si, cs:DskErr
                mov     cx, NumMsgs
LookUp:
                ;is this error message code equal to AH?
                cmp     ah, cs:[si]
                ;YES, lets go print it then
                je      Found
                ;NO, have we gone past it already?
                jnc     Unlisted
                ;NO, move pointer on to next error message code
                add     si, NextMsg
                loop    Lookup
Unlisted:
                ;we only get here if we can't find the error message
                ;set the error number of the unknown error
                lea     si, cs:UnkErc+15
                mov     al, ah
                call    Hex2
                lea     si, cs:UnkErc
Found:
                ;advance pointer by 1 to point to message
                inc     si
                ;display the error message
                call    WrtMsg
                ret
DspErr          endp

Hex2            proc    near
                push    ax
                mask_alhi
                hex2asc
                mov     cs:[si], al
                pop     ax
                mask_allo
                hex2asc
                mov     cs:[si+1], al
                ret
Hex2            endp

WrtMsg          proc    near
                push    si
                mov     cx, DriveLen
                lea     si, cs:Drive
                call    WrtChr
                pop     si
                mov     cx, MsgLen
WrtChr:
                movs    es:byte ptr[di], cs:[si]
                inc     di
                loop    WrtChr
                ret
WrtMsg          endp

;****************************************************************************
; MPLEX_INT handles interrupt 2FH.  If, on entry, AH is set to DSKWATCH's
; multiplex ID number, MPLEX_INT uses the value in AL as a function code.
; The functions supported are:
;
;    00H    Returns FFH in AL to indicate the program is installed
;           and returns the address of its signature in ES:DI.
;
;    01H    Report errors from floppy drives only.
;
;    02H    Report errors from hard drives only.
;
;    03H    Report errors on both hard drives and floppy drives.
;****************************************************************************

mplex_int       proc    far
                pushf                           ;Save FLAGS register
                cmp     ah, cs:[prog_id]        ;Branch if AH holds the
                je      mplex2                  ;  multiplex ID
                popf                            ;Restore FLAGS
                jmp     cs:[int2Fh]             ;Pass the interrupt on
;
; Function 00H verifies that the program is installed.
;
mplex2:         popf                            ;Restore FLAGS
                or      al, al                  ;Branch if function code
                jnz     mplex3                  ;  is other than 00H
                mov     al, 0FFh                ;Set AL to FFH
                push    cs                      ;Point ES:DI to the program
                pop     es                      ;  signature
                mov     di, offset Signature
                iret                            ;Return from interrupt
;
; Function 01h - floppy_drives
; Function 02h - hard_drives
; Function 03h - both_drives
;
mplex3:
                cmp     al, both_drives         ;if function > 3 then
                ja      mplex_exit              ;  mplex_exit
                mov     dw_drives, al           ;else set dw_drives
mplex_exit:
                iret                            ;Return from interrupt
mplex_int       endp

;****************************************************************************
; Data that will be discarded when the program becomes memory-resident.
;****************************************************************************


helpmsg         db      CR,LF
                db      "Monitors disk access and notifies you of any errors, unlike DOS",CR,LF
                db      "which will inform you after several consecutive errors.",CR,LF
                db      CR,LF
                db      "DSKWATCH [/?|/F|/H|/B]",CR,LF
                db      CR,LF
                db      "  /?  display this help screen.",CR,LF
                db      "  /F  display floppy disk errors ONLY.",CR,LF
                db      "  /H  display hard disk errors ONLY.",CR,LF
                db      "  /B  display both floppy disk and hard disk errors. (default)",CR,LF
                db      CR,LF
                db      "Author : Michael J Phillips",CR,LF
                db      "Date   : 09 Feb 93",CR,LF
                db      "Address: 5/5 Waddell Place",CR,LF
                db      "         Curtin ACT 2605",CR,LF
                db      "         Australia",CR,LF
                db      "Tel    : (06) 2811980",CR,LF
                db      CR,LF
                db      "FREEWARE!  This software may NOT be sold!",CR,LF
                db      "$"

errmsg1         db      "Syntax: DSKWATCH [/?|/F|/H|/B]",CR,LF,"$"
errmsg2         db      "Requires DOS 3.0 or higher",CR,LF,"$"
errmsg3         db      "Invalid parameter",CR,LF,"$"
errmsg4         db      "Program could not be installed",CR,LF,"$"

msg1            db      "DSKWATCH v3.04 - (c) 1987-1993 Vortex Software",CR,LF,"$"
msg2            db      "Monitoring both floppy and hard disk errors.",CR,LF,"$"
msg3            db      "Monitoring floppy disk errors only.",CR,LF,"$"
msg4            db      "Monitoring hard disk errors only.",CR,LF,"$"

parmflag        db      0                       ;0=No command line parameter
switch          db      ?


;****************************************************************************
; INIT makes the program resident in memory.
;****************************************************************************

                assume  cs:code,ds:code

init            proc    near
                mov     ah, 09h
                mov     dx, offset msg1
                int     21h

                mov     dw_drives, both_drives
                cld                             ;Clear direction flag
                mov     si, 81h                 ;Point SI to command line
                call    scanparam               ;Scan for "/" switch
                jnc     checkver                ;Branch if not found
                mov     switch, al
                cmp     al, '?'
                jne     chk1

                mov     ah, 09h                 ;Display help text
                mov     dx, offset helpmsg
                int     21h
                ;exit with errorlevel=0
                mov     ax, 4C00h
                int     21h
                ;***** EXIT *****
chk1:
                cmp     al, 'F'
                jne     chk2
                mov     dw_drives, floppy_drives
                jmp     checkver
chk2:
                cmp     al, 'H'
                jne     chk3
                mov     dw_drives, hard_drives
                jmp     checkver
chk3:
                cmp     al, 'B'
                jne     chk4
                mov     dw_drives, both_drives
                jmp     checkver
chk4:
                mov     dx, offset errmsg3
                mov     ah, 09h                 ;Display error message and
                int     21h                     ;  exit with ERRORLEVEL=1
;
; Check the DOS version and parse the command line.
;
checkver:       mov     dx, offset errmsg2      ;Exit if DOS version
                mov     ah, 30h                 ;  is less than 3.0
                int     21h
                cmp     al, 3
                jae     parse

error:          mov     ah, 09h                 ;Display error message
                int     21h
                ;exit with errorlevel=1
                mov     ax, 4C01h
                int     21h
                ;***** EXIT *****

parse:          call    findchar                ;Find parameter (if any)
                jc      checkprog               ;Branch if none found
                mov     parmflag, 1             ;Set PARMFLAG to 1

checkprog:      call    check_install           ;See if program is installed
                jnc     install                 ;Branch if it's not
;
; Communicate with an installed copy of the code.
;

                mov     al, dw_drives
                int     2Fh

                call    disp_param
cp2:
                ;exit with errorlevel=0
                mov     ax, 4C00h
                int     21h
                ;***** EXIT *****
;
; Install the program.
;
install:        call    mplex_id                ;Find a multiplex ID number
                mov     dx, offset errmsg4      ;Error if none available
                jc      error
                mov     prog_id, ah             ;Save the ID number

                mov     ax, 3513h               ;Hook interrupt 13H
                int     21h
                mov     word ptr int13h, bx
                mov     word ptr int13h[2], es
                mov     ax, 2513h
                mov     dx, offset disk_int
                int     21h

                mov     ax, 352Fh               ;Hook interrupt 2FH
                int     21h
                mov     word ptr int2Fh, bx
                mov     word ptr int2Fh[2], es
                mov     ax, 252Fh
                mov     dx, offset mplex_int
                int     21h

                ;calculate the screen
                mov     ah, INT10_GET_MODE
                int     10h
                sub     ah, DriveLen+MsgLen
                shl     ah, 1
                mov     byte ptr cs:VidOfs, ah
                test    al, 04h
                jne     NotColour
                add     cs:VidOfs, 08000h
NotColour:
                mov     ah, 49h                 ;Get the segment address of
                mov     es, ds:[2Ch]            ;  the environment block
                int     21h                     ;  and free the segment

                call    disp_param

                ;Terminate and Stay Resident
                mov     ax, 3100h               ;Terminate with function 31H
                mov     dx, (offset helpmsg - offset code + 15) shr 4
                int     21h
                ;***** EXIT *****

init            endp

disp_param      proc    near
                mov     al, dw_drives
                cmp     al, both_drives
                jne     dp1
                mov     dx, offset msg2
                jmp     dp3
dp1:
                cmp     al, floppy_drives
                jne     dp2
                mov     dx, offset msg3
                jmp     dp3
dp2:
                mov     dx, offset msg4
dp3:
                mov     ah, 09h
                int     21h
                ret
disp_param      endp

;****************************************************************************
; FINDCHAR advances SI to the next non-space or non-comma character.
; On return, carry set indicates EOL was encountered.
;****************************************************************************

findchar        proc    near
                lodsb                           ;Get the next character
                cmp     al, 20h                 ;Loop if space
                je      findchar
                cmp     al, 2Ch                 ;Loop if comma
                je      findchar
                dec     si                      ;Point SI to the character
                cmp     al, CR                  ;Exit with carry set if end
                je      eol                     ;  of line is reached

                clc                             ;Clear carry and exit
                ret

eol:            stc                             ;Set carry and exit
                ret
findchar        endp

;****************************************************************************
; SCANPARAM scans the command line for a / switch.  If found, carry returns
; set, SI contains its offset, and AL contain upcase(ch).  If not found,
; carry returns clear.
;****************************************************************************

scanparam       proc    near
                push    si                      ;Save SI
scanloop:       lodsb                           ;Get a character
                cmp     al, CR                  ;Exit if end of line
                je      scan_exit
                cmp     al, "/"                 ;Loop if not "/"
                jne     scanloop
                lodsb
                cmp     al, 'a'
                jb      skip
                cmp     al, 'z'
                ja      skip
                sub     al, 'a'-'A'
skip:
                mov     switch, al
                add     sp, 2                   ;Clear the stack
                stc                             ;Set carry and exit
                ret

scan_exit:      pop     si                      ;Restore SI
                clc                             ;Clear carry and exit
                ret
scanparam       endp

;****************************************************************************
; CHECK_INSTALL returns carry set if the program is already installed,
; carry clear if it's not.  If carry returns set, AH holds the program's
; multiplex ID number.
;****************************************************************************

check_install   proc    near
                mov     ax, 8000h               ;Initialize AH and AL
                mov     cx, 80h                 ;Initialize count

chinst1:        push    ax                      ;Save AX and CX
                push    cx
                sub     di, di                  ;Set ES and DI to 0
                mov     es, di
                int     2Fh                     ;Interrupt 2Fh
                cmp     al, 0FFh                ;Nothing here if AL isn't
                jne     chinst2                 ;  equal to FFH

                mov     si, offset Signature    ;See if program signature
                mov     cx, SignatureLen        ;  appears at the address
                repe    cmpsb                   ;  returned in ES:DI
                jne     chinst2                 ;Branch if it does not

                pop     cx                      ;Clear the stack and exit
                pop     ax                      ;  with carry set
                stc
                ret

chinst2:        pop     cx                      ;Retrieve AX and CX
                pop     ax
                inc     ah                      ;Next multiplex ID
                loop    chinst1                 ;Loop until done

                clc                             ;Exit with carry clear
                ret
check_install   endp

;****************************************************************************
; MPLEX_ID searches for an unused multiplex ID number.  If one is found,
; it is returned in AH with carry clear.  Carry set means no multiplex
; ID numbers are currently available.
;****************************************************************************

mplex_id        proc    near
                mov     ax, 8000h               ;Initialize AH and AL
                mov     cx, 80h                 ;Initialize count

mxid1:          push    ax                      ;Save AX and CX
                push    cx
                int     2Fh                     ;Interrupt 2Fh
                or      al, al                  ;Branch if AL=0
                jz      mxid2
                pop     cx                      ;Retrieve AX and CX
                pop     ax
                inc     ah                      ;Increment ID number
                loop    mxid1                   ;Loop until done

                stc                             ;Exit with carry set
                ret

mxid2:          pop     cx                      ;Clear the stack and exit
                pop     ax                      ;  with carry clear
                clc
                ret
mplex_id        endp

code            ends
                end     begin
